/* Fltkmchart.cxx */

/*
Copyright (c) 2015, The Fltkmchart Authors

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
*/

#include "Fltkmchart.h"

void Fltkmchart::drawLine_screenSpace(int x1, int y1, int x2, int y2, Fl_Color fc) {
	Fl_Color stash = fl_color();
	fl_color(fc);

	// change coordinate system - invert y positions
	fl_line(x1 + x(), y() + (h() - y1), x2 + x(), y() + (h() - y2));

	fl_color(stash);
}

void Fltkmchart::drawRect_screenSpace(int x, int y, int w, int h, Fl_Color fc) {
	Fl_Color stash = fl_color();
	fl_color(fc);

	// change coordinate system - invert y positions
	fl_rectf(this->x() + x, this->y() + (this->h() - (h + y)), w, h);

	fl_color(stash);
}

void Fltkmchart::drawText_screenSpace(const char* text, int x, int y, Fl_Color fc) {
	Fl_Font		stashface	= fl_font();
	Fl_Fontsize stashsize	= fl_size();
	Fl_Color	stashcol	= fl_color();

	fl_font(labelFontFace, labelFontSize);
	fl_color(fc);

	// change coordinate system - invert y positions
	fl_draw(text, this->x() + x, this->y() + (h() - y));

	fl_font(stashface, stashsize);
	fl_color(stashcol);
}

size_t Fltkmchart::getTextWidth(const char* text) {
	Fl_Font		stashface = fl_font();
	Fl_Fontsize stashsize = fl_size();

	fl_font(labelFontFace, labelFontSize);

	int x = 0;
	int y = 0;
	fl_measure(text, x, y);

	fl_font(stashface, stashsize);

	return x;
}

void Fltkmchart::drawLine_dataSpace(double x1, double y1, double x2, double y2, Fl_Color fc) {
	drawLine_screenSpace(xtf(x1), ytf(y1), xtf(x2), ytf(y2), fc);
}

void Fltkmchart::drawRect_dataSpace(double x1, double y1, double w, double h, Fl_Color fc) {
	drawRect_screenSpace(xtf(x1), ytf(y1), xtf(w), ytf(h), fc);
}

void Fltkmchart::clearSeries() {
	series.clear();

	// Recalculate x/y bounds
	x_min = 0;
	x_max = 1;
	y_min = 0;
	y_max = 1;

	// Recalculate transform functions
	recalculateTransformFunctions();
}

void Fltkmchart::addSeries(Fltkmchart_Series data) {
	series.push_back(data);

	// Recalculate x/y bounds
	for (size_t i = 0, e = data.points.size(); i != e; ++i) {
		x_min = std::min<double>(x_min, data.points[i].first);
		x_max = std::max<double>(x_max, data.points[i].first);
		y_min = std::min<double>(y_min, data.points[i].second);
		y_max = std::max<double>(y_max, data.points[i].second);
	}

	// Recalculate transform functions
	recalculateTransformFunctions();
}

std::function<std::string(double)> Fltkmchart::getNullAxisRenderer() {
	return [](double pt) {
		return std::string("");
	};
}

std::function<unsigned int(double)> Fltkmchart::getTransformFunction(double min, double max, unsigned int range, unsigned int padding) {	
	double scale = (range - padding) / (abs(max) + abs(min));

	return [range, padding, scale, min](double pt){
		double ret = ((pt - min) * scale) + padding;
		return static_cast<unsigned int>(round(ret));
	};
}

void Fltkmchart::recalculateTransformFunctions() {
	xtf = getTransformFunction(x_min, x_max, w(), xaxis_padding);
	ytf = getTransformFunction(y_min, y_max, h(), yaxis_padding);
}

void Fltkmchart::draw() {
	
	// Draw background
	drawRect_screenSpace(0, 0, w(), h(), backgroundColor);

	// Plot data points
	drawSeries();

	// Draw X, Y axis
	drawLine_dataSpace(0, y_min, 0, y_max, axisColor);
	drawLine_dataSpace(x_min, 0, x_max, 0, axisColor);
	
	// Draw axis labels
	std::string text;
	int fontheight = fl_height() - (2*fl_descent());

	auto ymark = ytf(0);
	for (double i = x_min; i < x_max; ++i) {
		text = xaxis_renderer(i);

		if (text.length()) {
			auto xmark = xtf(i);

			drawLine_screenSpace(xmark, ymark - 2, xmark, ymark + 2, axisColor);
			drawText_screenSpace(text.c_str(), xmark - (getTextWidth(text.c_str()) / 2), ymark - fontheight, axisColor);
		}
	}

	auto xmark = xtf(0);
	for (double i = y_min; i < y_max; ++i) {
		text = yaxis_renderer(i);
		if (text.length()) {
			auto ymark = ytf(i);

			drawLine_screenSpace(xmark - 2, ymark, xmark + 2, ymark, axisColor);
			drawText_screenSpace(text.c_str(), xmark - getTextWidth(text.c_str()) - 2, ymark - (fontheight/2), axisColor);
		}
	}

	// Draw legend
	for (size_t i = 0, e = series.size(); i != e; ++i) {
		auto ymark = h() - ((fontheight + 2) * i) - fontheight;
		auto yline = ymark + (fontheight / 2);
		
		drawLine_screenSpace(xaxis_padding + 2, yline,     xaxis_padding + 2 + 10, yline,     series[i].color);
		drawLine_screenSpace(xaxis_padding + 2, yline - 1, xaxis_padding + 2 + 10, yline - 1, series[i].color);
		drawText_screenSpace(series[i].legend.c_str(), xaxis_padding + 2 + 10 + 2, ymark, series[i].color);
	}

}

void Fltkmchart::resize(int x, int y, int w, int h) {
	recalculateTransformFunctions();
	Fl_Widget::resize(x, y, w, h); // super()
}

void Fltkmchart_Line::drawSeries() {
	
	for (size_t i = 0, e = series.size(); i != e; ++i) {

		for (size_t j = 1, je = series[i].points.size(); j < je; ++j) {

			double x = series[i].points[j].first;
			double y = series[i].points[j].second;
			double x_old = series[i].points[j - 1].first;
			double y_old = series[i].points[j - 1].second;

			drawLine_dataSpace(x_old, y_old, x, y, series[i].color);
		}

	}
}

void Fltkmchart_StaggeredLine::drawSeries() {

	for (size_t i = 0, e = series.size(); i != e; ++i) {

		for (size_t j = 1, je = series[i].points.size(); j < je; ++j) {

			double x = series[i].points[j].first;
			double y = series[i].points[j].second;
			double x_old = series[i].points[j - 1].first;
			double y_old = series[i].points[j - 1].second;

			drawLine_dataSpace(x_old, y_old, x, y_old, series[i].color);
			drawLine_dataSpace(x,     y_old, x, y,     series[i].color);
		}

	}
}

void Fltkmchart_Bar::drawSeries() {

	unsigned int y_0 = ytf(0);

	for (size_t i = 0, e = series.size(); i != e; ++i) {

		for (size_t j = 0, je = series[i].points.size(); j < je; ++j) {

			double x = series[i].points[j].first;
			double y = series[i].points[j].second;
			
			unsigned int x_1 = xtf(x + ( double(i)      / e));
			unsigned int x_2 = xtf(x + ((double(i) + 1) / e));

			drawRect_screenSpace(x_1, y_0, x_2, ytf(y), series[i].color);
		}

	}
}
