#include "FuelHistory.h"
#include <memory>
#include <map>
#include <ctime>

void FuelHistory::reset() {
	points.clear();
	times.clear();
	rates.clear();
	rates2.clear();
	first_odo = 0;
	first_time = 0;
}

void FuelHistory::loadData(std::vector<FuelRecord> record) {

	reset();

	double totalfuel = 0;
	double lastodo = 0;
	bool got_first = false;
	bool rr_valid = false;
	double rr_0;
	double rr_1;
	double rr_2;

	for (auto log : record) {

		if (!got_first) {
			first_odo = floor(log.odo / 1000) * 1000;
			first_time = log.date_time / 86400.0;
			got_first = true;
		}

		// Litres/km chart

		if (log.fuel_added == 0.0) {
			totalfuel = 0;
		} else {
			totalfuel += log.fuel_added;
		}

		if (log.odo == 0.0) {
			totalfuel = 0;
			points.push_back(std::make_pair(lastodo, 0));
			lastodo = 0;

		} else {
			lastodo = log.odo - first_odo;
			points.push_back(std::make_pair(lastodo, totalfuel));

		}

		// odo/time chart

		if (log.date_time != 0 && log.odo > 0) {
			times.push_back(std::make_pair(
				log.date_time / 86400 - first_time,
				log.odo - first_odo
				));
		}

		// consumption rate

		if (log.fuel_added == 0.0 || log.odo == 0.0) {
			rr_valid = false; // can't do anything with this data point
			rr_0 = 0;
			rr_1 = 0;
			rr_2 = 0;

		} else if (log.filled) {
			if (rr_valid) {
				double k_p_l = (log.odo - rr_1) / (log.fuel_added + rr_0);
				double l_p_100k = 100 / k_p_l;

				rates.push_back(std::make_pair(rr_2 / 86400.0 - first_time, k_p_l));
				rates.push_back(std::make_pair(log.date_time / 86400.0 - first_time, k_p_l));

				rates2.push_back(std::make_pair(rr_2 / 86400.0 - first_time, l_p_100k));
				rates2.push_back(std::make_pair(log.date_time / 86400.0 - first_time, l_p_100k));
			}

			rr_valid = true;
			rr_0 = 0;
			rr_1 = log.odo;
			rr_2 = static_cast<double>(log.date_time);

		} else { // filled = N
			if (rr_valid) {
				rr_0 += log.fuel_added;
			}

		}


	}
}

std::function<std::string(double)> FuelHistory::getSimpleAxisRenderer() {
	return [](double pt) {
		char buff[64] = { 0 };
		sprintf_s(buff, "%9.2f", pt);
		return std::string(buff);
	};
}

std::function<std::string(double)> FuelHistory::getDateAxisRenderer() {
	double first_time_ = this->first_time;

	return [first_time_](double pt) {
		time_t time = static_cast<time_t>((pt + first_time_) * 86400);

		struct tm *timespec = localtime(&time); // n.b. not thread safe

		if (timespec->tm_mday == 1) {
			char buff[64] = { 0 };
			sprintf_s(buff, "%04d-%02d", timespec->tm_year + 1900, timespec->tm_mon + 1);
			return std::string(buff);

		}

		return std::string();
	};
}

std::function<std::string(double)> FuelHistory::getOdoAxisRenderer() {
	double first_odo_ = this->first_odo;

	return [first_odo_](double pt) {
		if (static_cast<int>(pt) % 1000 == 0) {
			char buff[64] = { 0 };
			sprintf_s(buff, "%d", static_cast<int>(pt + first_odo_));
			return std::string(buff);

		} else {
			return std::string();
		}
	};
}

std::function<std::string(double)> FuelHistory::getDecimatedRenderer(int interval) {
	return [interval](double pt) {
		if (static_cast<int>(pt) % interval == 0) {
			char buff[64] = { 0 };
			sprintf_s(buff, "%d", static_cast<int>(pt));
			return std::string(buff);

		} else {
			return std::string();

		}
	};
}