/* main.cxx */

#include "flfuel.h"

#include <FL/fl_ask.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Native_File_Chooser.H>

#include "resource.h"

#include <vector>
#include <string>
#include <ctime>

#include "FuelRecord.h"
#include "FuelHistory.h"

static Fl_Double_Window *mainwindow = nullptr, *popupwindow = nullptr;
static Fl_Text_Buffer *ftbsummary = new Fl_Text_Buffer();

static std::string currentFileName = "";
static std::vector<FuelRecord> data;

static size_t editrecord = -1; // FFFFF...

void refreshviews() {
	char buff[4096] = { 0 };

	//

	FuelHistory fh;
	fh.loadData(data);

	// 

	ftbsummary->remove(0, ftbsummary->length());

	sprintf_s(buff, "Current file: %s\n", (currentFileName.length() ? currentFileName.c_str() : "(new file)"));
	ftbsummary->append(buff);

	sprintf_s(buff, "Data points: %d\n", data.size());
	ftbsummary->append(buff);

	if (fh.rates.size()) {
		sprintf_s(buff, "Last efficiency: %1.1f K/L (%1.1f L/100Km)\n",
			fh.rates [fh.rates. size() - 1].second,
			fh.rates2[fh.rates2.size() - 1].second
		);
		ftbsummary->append(buff);
	}

	txSummary->redraw();

	//

	datatree->clear();


	for (size_t i = 0, e = data.size(); i != e; ++i) {
		
		Fl_Tree_Item *node = datatree->add("_unused_name");
		node->label(data[i].date.c_str());

		if (data[i].odo == 0.0) {
			Fl_Tree_Item *last = datatree->add(node, "[!] Odometer reading UNKNOWN");
			last->labelcolor(FL_RED);
		} else {
			sprintf_s(buff, "Odometer reading: %1.0f Km", data[i].odo);
			datatree->add(node, buff);
		}

		if (data[i].fuel_added == 0.0) {
			Fl_Tree_Item *last = datatree->add(node, "[!] Fuel added: UNKNOWN");
			last->labelcolor(FL_RED);
		} else {
			sprintf_s(buff, "Fuel added: %1.2f Litres", data[i].fuel_added);
			datatree->add(node, buff);
		}
		
		if (!data[i].filled) {
			datatree->add(node, "[!] Not a full tank");
		}

		if (data[i].comment.length()) {
			sprintf_s(buff, "%s", data[i].comment.c_str());
			Fl_Tree_Item *last = datatree->add(node, buff);
			last->labelcolor(FL_GRAY);
		}

	}

	datatree->root_label("Record History");
	datatree->redraw();
	

	//
	
	{
		Fltkmchart_Series d;
		d.color = FL_RED;
		d.legend = "Km / Time";
		d.points = fh.times;

		chart_disttime->clearSeries();
		chart_disttime->addSeries(d);
		chart_disttime->setXAxisRenderer(fh.getDateAxisRenderer());
		chart_disttime->setYAxisRenderer(fh.getOdoAxisRenderer());

		chart_disttime->redraw();
	}

	{
		Fltkmchart_Series d;
		d.color = FL_RED;
		d.legend = "Litres / Km";
		d.points = fh.points;

		chart_usedist->clearSeries();
		chart_usedist->addSeries(d);
		chart_usedist->setXAxisRenderer(fh.getOdoAxisRenderer());
		chart_usedist->setYAxisRenderer(fh.getDecimatedRenderer(40));

		chart_usedist->redraw();
	}

	{
		Fltkmchart_Series kpl;
		kpl.color = FL_RED;
		kpl.legend = "Km / Litre (higher is better)";
		kpl.points = fh.rates;

		Fltkmchart_Series lhk;
		lhk.color = FL_BLUE;
		lhk.legend = "Litres / 100 Km (lower is better)";
		lhk.points = fh.rates2;

		chart_kpl->clearSeries();
		chart_kpl->addSeries(kpl);
		chart_kpl->addSeries(lhk);
		chart_kpl->setXAxisRenderer(fh.getDateAxisRenderer());
		chart_kpl->setYAxisRenderer(fh.getSimpleAxisRenderer());

		chart_kpl->redraw();
	}

}

void main_onNewFile() {
	currentFileName = "";
	data.clear();
	refreshviews();
}

void openFileByName(const char* filename) {

	FILE* fh = fopen(filename, "rb");
	if (!fh) {
		fl_alert("Couldn't open file '%s'!", filename);
		return;
	}

	main_onNewFile();
	currentFileName = filename;
	for (;;) {
		FuelRecord f;
		if (!f.deserialise_from(fh)) {
			break;
		}
		data.push_back(f);
	}

	fclose(fh);

	refreshviews();
}

void main_onOpenFile() {
	Fl_Native_File_Chooser dlg;
	
	dlg.title("Open file...");
	dlg.type(Fl_Native_File_Chooser::BROWSE_FILE);
	dlg.filter("FLFuel catalogs\t*.flf\n");

	// Show native chooser
	switch (dlg.show()) {
		case -1: // error
		case 1:  // cancelled
			return;		
	}

	openFileByName(dlg.filename());
}

void main_onSaveFile() {

	// 'save as' handling
	if (! currentFileName.length()) {
		Fl_Native_File_Chooser dlg;

		dlg.title("Save as...");
		dlg.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
		dlg.filter("FLFuel catalogs\t*.flf\n");

		switch (dlg.show()) {
			case -1: // error
			case 1:  // cancelled
				return; 
			default:
				currentFileName = dlg.filename();
		}
	}

	// open file
	FILE* fh = fopen(currentFileName.c_str(), "wb");
	if (!fh) {
		fl_alert("Couldn't open file '%s'!", currentFileName.c_str());
		return;
	}

	// serialise+write
	for (size_t i = 0, e = data.size(); i != e; ++i) {
		auto line = data[i].serialise();
		fwrite(line.c_str(), line.length(), 1, fh);
	}

	// done
	fclose(fh);
}

void main_onExit() {
	mainwindow->hide();
}

void main_onAddClicked() {
	editrecord = -1;
	
	char buff[64] = { 0 };
	time_t curtime = time(NULL);
	struct tm* timespec = localtime(&curtime);
	sprintf_s(buff, "%04d-%02d-%02d", timespec->tm_year + 1900, timespec->tm_mon + 1, timespec->tm_mday);

	recfield_date->value(buff);
	recfield_odo->value(0.0);
	recfield_added->value(0.0);
	recfield_filled->value(1);
	recfield_comment->value("");	

	popupwindow->show();
}

void tryGetSelected() {
	editrecord = -1; // not found

	if (!datatree->root()) {
		// No root node = no children ever added
		return;
	}

	auto *cur = datatree->get_item_focus();
	if (!cur || !cur->parent()) {
		return;
	}

	if (cur->parent() == datatree->root()) {
		// ok

	} else if (cur->parent()->parent() == datatree->root()) {
		cur = cur->parent();

	} else {
		// Not nested the way we expected
		return;
	}

	// Find record index
	for (int i = 0, e = datatree->root()->children(); i != e; ++i) {
		if ((Fl_Tree_Item*)(datatree->root()->child(i)) == cur) {
			editrecord = i;
			return;
		}
	}

	// Not found
}

void main_onDeleteClicked() {
	tryGetSelected();

	if (editrecord == -1) {
		fl_alert("No entry selected.");
		return;
	}

	if (fl_ask("Are you sure you want to remove this entry?")) {
		auto itr = data.begin() + editrecord;
		data.erase(itr);
		refreshviews();
	}
}

void main_onEditClicked() {
	tryGetSelected();

	if (editrecord == -1) {
		fl_alert("No entry selected.");
		return;
	}

	recfield_date->value(data[editrecord].date.c_str());
	recfield_odo->value(data[editrecord].odo);
	recfield_added->value(data[editrecord].fuel_added);
	recfield_filled->value(data[editrecord].filled ? 1 : 0);
	recfield_comment->value(data[editrecord].comment.c_str());

	popupwindow->show();
}

void main_onInputSubmitted() {
	if (popupwindow == nullptr) {
		return;
	}

	FuelRecord f;
	f.date       = recfield_date->value();
	f.odo        = recfield_odo->value();
	f.fuel_added = recfield_added->value();
	f.comment    = recfield_comment->value();
	f.filled	 = (recfield_filled->value() == 1);
	
	std::string error = f.validate();
	if (error.length()) {
		fl_alert("Can't save record: %s", error.c_str());
		return;
	}

	if (editrecord == -1) {
		data.push_back(f);
	} else {
		data[editrecord] = f; // replace
		editrecord = -1;
	}

	popupwindow->hide();

	refreshviews();
}

void main_onCancelClicked() {
	popupwindow->hide();
}

Fl_Font getfontbyname(const char* name) {
	size_t  name_len    = strlen(name);
	Fl_Font total_fonts = Fl::set_fonts();

	for (Fl_Font i = 0; i < total_fonts; ++i) {
		if (!memcmp(name, Fl::get_font_name(i), name_len)) {
			return i;
		}
	}

	return -1;
}

bool setfldefaultfont(const char* name, Fl_Fontsize size) {

	FL_NORMAL_SIZE = size; // swap the default font size

	Fl_Font f = getfontbyname(name);
	if (f != -1) {
		Fl::set_font(0, f); // swap the default font ptr
		return true;
	} else {
		return false;
	}

}

int main(int argc, char** argv) {

#ifdef WIN32
	setfldefaultfont("Segoe UI", 11) || setfldefaultfont("Tahoma", 11);
#endif

	Fl::scheme("gtk+");

	mainwindow = makewindow_main();

	popupwindow = makewindow_inputpoint();
	popupwindow->set_modal();

#ifdef WIN32
	mainwindow->icon(LoadIcon(fl_display, MAKEINTRESOURCE(IDI_ICON1)));
#endif

	mainwindow->show();
	
	txSummary->buffer(ftbsummary);
	
	refreshviews();

	if (argc >= 2) {
		openFileByName(argv[1]);
	}
	
	return Fl::run();
}
